# #
# @copyright Copyright (C) 2023 SpyroSoft Solutions S.A. All rights reserved.
# #

import platform
import subprocess
from pathlib import Path
from typing import Optional

from scargo.global_values import SCARGO_DEFAULT_CONFIG_FILE, SCARGO_LOCK_FILE


def find_program_path(program_name: str) -> Optional[Path]:
    cmd = "where" if platform.system() == "Windows" else "/usr/bin/which"
    try:
        cmd_output = subprocess.check_output([cmd, program_name])
    except subprocess.CalledProcessError:
        return None
    return Path(cmd_output.decode("utf-8").strip())


def get_config_file_path(config_file_name: str) -> Optional[Path]:
    current_path = Path.cwd()
    directories_to_check = [current_path] + list(current_path.parents)
    for directory in directories_to_check:
        if (directory / config_file_name).exists():
            return directory / config_file_name
    return None


def get_project_root_or_none() -> Optional[Path]:
    config_path = get_config_file_path(SCARGO_LOCK_FILE) or get_config_file_path(
        SCARGO_DEFAULT_CONFIG_FILE
    )
    return config_path.parent if config_path else None



# #
# @copyright Copyright (C) 2023 SpyroSoft Solutions S.A. All rights reserved.
# #

"""Handle docker for project"""
import shutil
import subprocess
import sys
from pathlib import Path
from typing import List, Optional, Sequence

import docker

from scargo.config_utils import get_scargo_config_or_exit
from scargo.logger import get_logger

logger = get_logger()


def scargo_docker_build(
    docker_opts: Sequence[str], project_root: Optional[Path] = None
) -> None:
    """
    Build docker

    :param docker_opts: additional docker options
    :param project_root
    :raises CalledProcessError: if docker build fail
    """
    logger.debug("Build docker environment.")
    if not project_root:
        project_root = get_scargo_config_or_exit().project_root
    docker_path = _get_docker_path(project_root)

    cmd = get_docker_compose_command()
    cmd.extend(["build", *docker_opts])

    try:
        subprocess.run(cmd, cwd=docker_path, check=True)
        logger.info("Initialize docker environment.")
    except subprocess.CalledProcessError:
        logger.error("Build docker fail.")
        sys.exit(1)


def scargo_docker_run(
    docker_opts: Sequence[str],
    command: Optional[str] = None,
) -> None:
    """
    Run docker

    :param docker_opts: additional docker options
    :param command: command to run in the container
    :raises CalledProcessError: if docker did not start
    """
    logger.debug("Run docker environment.")

    config = get_scargo_config_or_exit()
    docker_path = _get_docker_path(config.project_root)
    project_config_name = config.project.name

    cmd = get_docker_compose_command()
    cmd.extend(
        [
            "run",
            *docker_opts,
            f"{project_config_name}_dev",
        ]
    )
    if command:
        cmd.extend(["bash", "-c", command])

    try:
        subprocess.run(cmd, cwd=docker_path, check=True)
        logger.info("Stop docker environment.")
    except subprocess.CalledProcessError:
        sys.exit(1)


def scargo_docker_exec(docker_opts: List[str]) -> None:
    """
    Exec docker

    :param docker_opts: additional docker options
    :raises CalledProcessError: if docker did not start
    """
    logger.debug("Exec docker environment.")

    config = get_scargo_config_or_exit()
    image = config.project.docker_image_tag

    if not image:
        logger.error("docker-image-tag not defined in .toml under project section")
        sys.exit(1)

    client = docker.from_env()
    newest_container = client.containers.list(
        limit=1, filters={"ancestor": image, "status": "running"}
    )
    if not newest_container:
        logger.error("No running containers using image `%s` to attach to!", image)
        logger.info("Use scargo docker run to run container.")
        sys.exit(1)

    bash_command = ["bash"]
    cmd = ["docker", "exec", "-it", *docker_opts, newest_container[0].id, *bash_command]
    try:
        subprocess.run(cmd, check=True)
        logger.info("Stop exec docker environment.")
    except subprocess.CalledProcessError:
        sys.exit(1)


def _get_docker_path(project_path: Path) -> Path:
    # do not rebuild dockers in the docker
    if Path("/.dockerenv").exists():
        logger.error("Cannot used docker command inside the docker container.")
        sys.exit(1)
    return Path(project_path, ".devcontainer")


def get_docker_compose_command() -> List[str]:
    """Get docker command

    Returns:
        List[str]: _description_
    """
    command = ["docker-compose"]
    # Check if docker-compose or docker compose is available
    if shutil.which("docker-compose"):
        command = ["docker-compose"]
    elif shutil.which("docker"):
        command = ["docker", "compose"]
    else:
        logger.error("Neither docker-compose nor docker compose are available.")
        sys.exit(1)

    return command



# #
# @copyright Copyright (C) 2023 SpyroSoft Solutions S.A. All rights reserved.
# #
"""Check written code with formatters"""
import abc
import glob
import os
import re
import subprocess
import sys
from itertools import chain
from pathlib import Path
from typing import Iterable, List, NamedTuple, Sequence, Type

from scargo.clang_utils import get_comment_lines
from scargo.config import CheckConfig, Config, TodoCheckConfig
from scargo.config_utils import prepare_config
from scargo.logger import get_logger

logger = get_logger()


def scargo_check(
    clang_format: bool,
    clang_tidy: bool,
    copy_right: bool,
    cppcheck: bool,
    cyclomatic: bool,
    pragma: bool,
    todo: bool,
    verbose: bool,
) -> None:
    """
    Check written code using different formatters

    :param bool clang_format: check clang_format
    :param bool clang_tidy: check clang_tidy
    :param bool copy_right:  check copyrights
    :param bool cppcheck: check cpp format
    :param bool cyclomatic: check cyclomatic
    :param bool pragma: check pragma
    :param bool todo: check todo left in code
    :param bool verbose: set verbose
    :return: None
    """
    config = prepare_config()

    # Todo, remove chdir and change cwd for checks
    os.chdir(config.project_root)

    checkers: List[Type[CheckerFixer]] = []
    if clang_format:
        checkers.append(ClangFormatChecker)
    if clang_tidy:
        checkers.append(ClangTidyChecker)
    if copy_right:
        checkers.append(CopyrightChecker)
    if cppcheck:
        checkers.append(CppcheckChecker)
    if cyclomatic:
        checkers.append(CyclomaticChecker)
    if pragma:
        checkers.append(PragmaChecker)
    if todo:
        checkers.append(TodoChecker)

    # Run all checks by default
    if not checkers:
        checkers = [
            ClangFormatChecker,
            ClangTidyChecker,
            CopyrightChecker,
            CppcheckChecker,
            CyclomaticChecker,
            PragmaChecker,
            TodoChecker,
        ]

    problem_counts = []
    for checker_class in checkers:
        problem_count = checker_class(config, verbose=verbose).check()
        problem_counts.append((checker_class, problem_count))
    if len(checkers) > 1:
        logger.info("Summary:")
        if any(count > 0 for _, count in problem_counts):
            for checker_class, problem_count in problem_counts:
                logger.info(
                    f"{checker_class.check_name}: {problem_count} problems found"
                )
            sys.exit(1)
        else:
            logger.info("No problems found!")


class CheckResult(NamedTuple):
    problems_found: int
    fix: bool = True


class CheckerFixer(abc.ABC):
    check_name: str
    headers_only = False
    can_fix = False

    def __init__(
        self, config: Config, fix_errors: bool = False, verbose: bool = False
    ) -> None:
        self._config = config
        self._fix_errors = fix_errors
        self._verbose = verbose

    def check(self) -> int:
        logger.info(f"Starting {self.check_name} check...")
        error_count = self.check_files()
        self.report(error_count)
        return error_count

    def check_files(self) -> int:
        error_counter = 0
        for file_path in find_files(
            self._config.source_dir_path,
            ("*.h", "*.hpp") if self.headers_only else ("*.h", "*.hpp", "*.c", "*.cpp"),
            self.get_exclude_patterns(),
        ):
            result = self.check_file(file_path)
            error_counter += result.problems_found
            if (
                result.problems_found > 0
                and self._fix_errors
                and self.can_fix
                and result.fix
            ):
                logger.info("Fixing...")
                self.fix_file(file_path)
        return error_counter

    def report(self, count: int) -> None:
        problem_count = self.format_problem_count(count)
        if self._fix_errors and self.can_fix:
            logger.info(f"Finished {self.check_name} check. Fixed {problem_count}.")
        else:
            logger.info(f"Finished {self.check_name} check. Found {problem_count}.")
            if count > 0:
                logger.error(f"{self.check_name} check fail!")

    @staticmethod
    def format_problem_count(count: int) -> str:
        return f"problems in {count} files"

    def get_exclude_patterns(self) -> List[str]:
        return [*self._config.check.exclude, *self.get_check_config().exclude]

    def get_check_config(self) -> CheckConfig:
        return getattr(  # type: ignore[no-any-return]
            self._config.check, self.check_name.replace("-", "_")
        )

    @abc.abstractmethod
    def check_file(self, file_path: Path) -> CheckResult:
        pass

    def fix_file(self, file_path: Path) -> None:
        pass


class PragmaChecker(CheckerFixer):
    check_name = "pragma"
    headers_only = True
    can_fix = True

    def check_file(self, file_path: Path) -> CheckResult:
        with open(file_path, encoding="utf-8") as file:
            for line in file.readlines():
                if "#pragma once" in line:
                    return CheckResult(0)
        logger.warning("Missing '#pragma once' in %s", file_path)
        return CheckResult(1)

    def fix_file(self, file_path: Path) -> None:
        with open(file_path, encoding="utf-8") as file:
            old = file.read()

        with open(file_path, "w", encoding="utf-8") as file:
            file.write("#pragma once\n")
            file.write("\n")
            file.write(old)


class CopyrightChecker(CheckerFixer):
    check_name = "copyright"
    can_fix = True

    def __init__(self, config: Config, fix_errors: bool = False, verbose: bool = False):
        super().__init__(config, fix_errors, verbose)
        self.copyright_desc = self.get_check_config().description or ""

    def check(self) -> int:
        if not self.copyright_desc:
            logger.warning(
                "No copyright line defined in scargo.toml at check.copyright.description"
            )
            return 0
        return super().check()

    def __get_copyright_lines(self) -> List[str]:
        return (
            ["//\n"]
            + [f"// {el}\n" for el in self.copyright_desc.split("\n")]
            + ["//\n"]
        )

    def check_file(self, file_path: Path) -> CheckResult:
        copyright_lines = self.__get_copyright_lines()

        with open(file_path, encoding="utf-8") as file:
            file_lines = file.readlines()

            for line_no, line in enumerate(file_lines[: -(len(copyright_lines) - 1)]):
                if line == copyright_lines[0]:
                    if all(
                        line_from_file == line_from_copyrights
                        for line_from_file, line_from_copyrights in zip(
                            file_lines[line_no:], copyright_lines
                        )
                    ):
                        return CheckResult(problems_found=0)

        logger.warning("Missing copyright line in %s.", file_path)
        return CheckResult(problems_found=1)

    def fix_file(self, file_path: Path) -> None:
        with open(file_path, encoding="utf-8") as file:
            old = file.read()

        with open(file_path, "w", encoding="utf-8") as file:
            for line in self.__get_copyright_lines():
                file.write(line)
            file.write("\n")
            file.write(old)


class TodoChecker(CheckerFixer):
    check_name = "todo"

    @staticmethod
    def format_problem_count(count: int) -> str:
        return f"{count} problems"

    def check_file(self, file_path: Path) -> CheckResult:
        keywords = self.get_check_config().keywords
        keyword_patterns = [
            re.compile(rf"\b{re.escape(keyword)}\b") for keyword in keywords
        ]
        error_counter = 0
        for line_number, line in get_comment_lines(file_path):
            for keyword, keyword_pattern in zip(keywords, keyword_patterns):
                if keyword_pattern.search(line):
                    error_counter += 1
                    logger.warning(
                        f"Found {keyword} in {file_path} at line {line_number}"
                    )
        return CheckResult(error_counter)

    def get_check_config(self) -> TodoCheckConfig:
        return self._config.check.todo


class ClangFormatChecker(CheckerFixer):
    check_name = "clang-format"
    can_fix = True

    def check_file(self, file_path: Path) -> CheckResult:
        cmd = [
            "/usr/bin/clang-format",
            "--style=file",
            "--dry-run",
            "-Werror",
            str(file_path),
        ]
        try:
            subprocess.check_output(cmd)
        except subprocess.CalledProcessError as e:
            if self._verbose:
                logger.info(e.output.decode())
            else:
                logger.warning("clang-format found error in file %s", file_path)
            return CheckResult(1)
        return CheckResult(0)

    def fix_file(self, file_path: Path) -> None:
        subprocess.check_call(
            ["/usr/bin/clang-format", "-style=file", "-i", str(file_path)]
        )


class ClangTidyChecker(CheckerFixer):
    check_name = "clang-tidy"
    build_path = Path("./build/")

    def check_file(self, file_path: Path) -> CheckResult:
        cmd = ["clang-tidy", str(file_path)]
        if self._config.project.target.family == "esp32":
            cmd.extend(["-p", str(self.build_path)])
            if not Path(self.build_path, "compile_commands.json").exists():
                # creates compilation database and runs run-clang-tidy.py on the whole project
                # (the latter is not needed, but there's no option to suppress it)
                subprocess.run(
                    ["idf.py", "clang-check"],
                    stdout=subprocess.DEVNULL,
                    stderr=subprocess.DEVNULL,
                    check=False,
                )
        if file_path.suffix == ".h":
            cmd.extend(["--", "-x", "c++"])
        if "--" not in cmd and self._config.project.target.family != "esp32":
            cmd.append("--")
        try:
            subprocess.check_output(cmd)
        except subprocess.CalledProcessError as e:
            if self._verbose:
                logger.info(e.output.decode())
            else:
                logger.warning("clang-tidy found error in file %s", file_path)
            return CheckResult(1)
        return CheckResult(0)


def find_files(
    dir_path: Path, glob_patterns: Sequence[str], exclude_patterns: Sequence[str]
) -> Iterable[Path]:
    exclude_list = [path for pattern in exclude_patterns for path in glob.glob(pattern)]

    for file_path in chain.from_iterable(
        dir_path.rglob(pattern) for pattern in glob_patterns
    ):
        if file_path.is_file():
            if any(exclude in str(file_path) for exclude in exclude_list):
                logger.info("Skipping %s", file_path)
                continue

            yield file_path


class CyclomaticChecker(CheckerFixer):
    check_name = "cyclomatic"

    def check_files(self) -> int:
        cmd = ["lizard", str(self._config.source_dir_path), "-C", "25", "-w"]

        for exclude_pattern in self.get_exclude_patterns():
            cmd.extend(["--exclude", exclude_pattern])
        try:
            subprocess.check_call(cmd)
        except subprocess.CalledProcessError:
            logger.error(f"{self.check_name} fail!")
        return 0

    def report(self, count: int) -> None:
        logger.info(f"Finished {self.check_name} check.")

    def check_file(self, file_path: Path) -> CheckResult:
        raise NotImplementedError


class CppcheckChecker(CheckerFixer):
    check_name = "cppcheck"

    def check_files(self) -> int:
        cmd = "cppcheck --enable=all --suppress=missingIncludeSystem --inline-suppr src/ main/"
        try:
            subprocess.check_call(cmd, shell=True)
        except subprocess.CalledProcessError:
            logger.error(f"{self.check_name} fail!")
        return 0

    def report(self, count: int) -> None:
        logger.info(f"Finished {self.check_name} check.")

    def check_file(self, file_path: Path) -> CheckResult:
        raise NotImplementedError



import sys
from pathlib import Path, PurePosixPath
from typing import List

import docker as dock
from docker import DockerClient

from scargo.config import ProjectConfig
from scargo.global_values import SCARGO_DOCKER_ENV
from scargo.logger import get_logger

logger = get_logger()


def run_scargo_again_in_docker(
    project_config: ProjectConfig, project_path: Path
) -> None:
    """
    Run command in docker

    :param dict project_config: project configuration
    :param Path project_path: path to project root
    :return: None
    """
    build_env = project_config.build_env
    if build_env != SCARGO_DOCKER_ENV or Path("/.dockerenv").exists():
        return

    relative_path = Path.cwd().relative_to(project_path)
    path_in_docker = PurePosixPath("/workspace", relative_path)

    cmd_args = sys.argv[1:]
    try:
        # Correct base-dir arg for command in docker
        for idx, val in enumerate(cmd_args):
            if val == "-B" or val == "--base-dir":
                cmd_args[idx + 1] = "."
    except Exception as e:
        logger.error(e)

    entrypoint = ""
    if project_config.target.family == "esp32":
        entrypoint = "/opt/esp/entrypoint.sh"

    docker_tag = project_config.docker_image_tag
    client = dock.from_env()

    status_code = run_command_in_docker(
        ["scargo", *cmd_args],
        client,
        docker_tag,
        entrypoint,
        project_path,
        path_in_docker,
    )
    sys.exit(status_code)


def run_command_in_docker(  # type: ignore[no-any-unimported]
    command: List[str],
    client: DockerClient,
    docker_tag: str,
    entrypoint: str,
    project_path: Path,
    path_in_docker: PurePosixPath,
) -> int:
    logger.info(f"Running '{' '.join(command)}' command in docker.")
    container = client.containers.run(
        docker_tag,
        command,
        volumes=[f"{project_path}:/workspace/", "/dev/:/dev/"],
        entrypoint=entrypoint,
        privileged=True,
        detach=True,
        working_dir=str(path_in_docker),
    )
    output = container.attach(stdout=True, stream=True, logs=True, stderr=True)
    for line in output:
        print(line.decode(), end="")
    result = container.wait()
    container.remove()
    return result["StatusCode"]  # type: ignore[no-any-return]



# #
# @copyright Copyright (C) 2023 SpyroSoft Solutions S.A. All rights reserved.
# #
from typing import List, Sequence


class ArgumentDescriptor:
    """Describes a function argument"""

    def __init__(self, name: str, data_type: str):
        self.name = name
        self.data_type = data_type


class FunctionDescriptor:
    """Contains function name and types"""

    def __init__(
        self,
        name: str,
        return_type: str,
        specifiers: Sequence[str],
        arguments: Sequence[ArgumentDescriptor] = (),
    ):
        self.name = name
        self.return_type = return_type
        self.specifiers = specifiers
        self.arguments = arguments

        for _, arg in enumerate(self.arguments):
            if not arg.name:
                arg.name = ""

    def get_specifiers(self) -> str:
        return ", ".join(spec for spec in self.specifiers)

    def get_typed_args(self) -> str:
        for arg in self.arguments:
            if arg.data_type == "void":
                return " "
        return ", ".join(arg.data_type + " " + arg.name for arg in self.arguments)

    def get_arg_names(self) -> str:
        return ", ".join(arg.name for arg in self.arguments)

    def get_arg_types(self) -> str:
        for arg in self.arguments:
            if arg.data_type == "void":
                return " "
        return ", ".join(arg.data_type for arg in self.arguments)


class ClassDescriptor:
    """Contains class names and function definitions"""

    def __init__(self, name: str, mock_name: str, methods: List[FunctionDescriptor]):
        self.name = name
        self.mock_name = mock_name
        self.methods: List[FunctionDescriptor] = methods


class NamespaceDescriptor:
    def __init__(self, name: str):
        self.name = name


class IncludeDescriptor:
    def __init__(self, name: str):
        self.name = name


class HeaderDescriptor:
    """Parsed header definitions"""

    def __init__(
        self,
        name: str,
        classes: List[ClassDescriptor],
        namespaces: List[NamespaceDescriptor],
        includes: List[IncludeDescriptor],
    ):
        self.name = name
        self.classes = classes
        self.namespaces = namespaces
        self.includes = includes



#!/usr/bin/env python3
# #
# @copyright Copyright (C) 2023 SpyroSoft Solutions S.A. All rights reserved.
# #

from pathlib import Path

from scargo.config import Config
from scargo.file_generators.base_gen import create_file_from_template
from scargo.file_generators.clang_parser.header_parser import parse_file

MOCKS_DIR = "tests/mocks"


def generate_mocks(src_header: Path, config: Config) -> bool:
    """
    Generates mock header and implementations for specified source headers.
    Creates directories and CMake lists where required.
    :param src_header Path to source directory or header
    :param config
    """

    dst_header = get_mock_path(src_header, config)

    # Skip generation if destination header exists
    if dst_header.is_file():
        return False

    template_and_output_paths = [
        ("mock/class_interface.h.j2", dst_header),
        ("mock/class_mock.h.j2", dst_header.with_name(f"mock_{dst_header.name}")),
        ("mock/class_mock.cpp.j2", dst_header.with_name(f"mock_{dst_header.stem}.cpp")),
    ]

    # Read classes and functions from the source header
    header_data = parse_file(src_header)

    for template_path, output_path in template_and_output_paths:
        create_file_from_template(
            template_path,
            output_path,
            template_params={"header": header_data},
            config=config,
        )

    return True


def get_mock_path(header_path: Path, config: Config) -> Path:
    header_path_from_src = header_path.relative_to(config.source_dir_path)
    return config.project_root / MOCKS_DIR / header_path_from_src



# #
# @copyright Copyright (C) 2023 SpyroSoft Solutions S.A. All rights reserved.
# #

import os
import re
import shutil
import subprocess
import sys
from pathlib import Path
from shutil import copyfile
from typing import Optional

from scargo.config import Config
from scargo.config_utils import prepare_config
from scargo.file_generators.mock_gen import generate_mocks
from scargo.file_generators.ut_gen import generate_ut
from scargo.global_values import SCARGO_PKG_PATH
from scargo.logger import get_logger

logger = get_logger()

OUT_FS_DIR = Path("build", "fs")


def scargo_gen(
    profile: str,
    gen_ut: Optional[Path],
    gen_mock: Optional[Path],
    certs: Optional[str],
    certs_mode: Optional[str],
    certs_input: Optional[Path],
    certs_passwd: Optional[str],
    fs: bool,
    single_bin: bool,
) -> None:
    config = prepare_config()

    if gen_ut:
        generate_ut(gen_ut, config)

    if gen_mock:
        if gen_mock.suffix not in (".h", ".hpp"):
            logger.error("Not a header file. Please chose .h or .hpp file.")
            sys.exit(1)
        if generate_mocks(gen_mock, config):
            logger.info(f"Generated: {gen_mock}")
        else:
            logger.info(f"Skipping: {gen_mock}")

    if certs:
        generate_certs(
            certs, certs_mode, certs_input, certs_passwd, config.project_root
        )

    if fs:
        generate_fs(config)

    if single_bin:
        project_profile_path = config.project_root / "build" / profile
        gen_single_binary(project_profile_path, config)


def generate_certs(
    device_name: str,
    mode_for_certs: Optional[str],
    certs_intermediate_dir: Optional[Path],
    certs_passwd: Optional[str],
    project_path: Path,
) -> None:
    internal_certs_dir = Path(SCARGO_PKG_PATH, "certs")
    projects_builds_path = project_path / "build"
    certs_out_dir = projects_builds_path / "certs"
    if not certs_intermediate_dir:
        certs_intermediate_dir = projects_builds_path / "certs"
    if not certs_passwd:
        certs_passwd = "1234"

    if mode_for_certs == "all":
        mode_for_certs = "All-certificates"
    elif mode_for_certs == "device":
        mode_for_certs = "Device-certificate"
    else:
        mode_for_certs = "All-certificates"

    certs_out_dir.mkdir(parents=True, exist_ok=True)
    subprocess.call(
        [
            internal_certs_dir / "generateAllCertificates.sh",
            "--name",
            device_name,
            "--mode",
            mode_for_certs,
            "--output",
            certs_out_dir,
            "--input",
            certs_intermediate_dir,
            "--passwd",
            certs_passwd,
        ]
    )

    certs_out_azure_dir = certs_out_dir / "azure"
    certs_out_uc_dir = project_path / OUT_FS_DIR

    certs_out_uc_dir.mkdir(parents=True, exist_ok=True)
    certs_out_azure_dir.mkdir(parents=True, exist_ok=True)

    try:
        copyfile(
            certs_out_dir / "certs" / "iot-device.cert.pem",
            certs_out_uc_dir / "device_cert.pem",
        )
        copyfile(
            certs_out_dir / "private" / "iot-device.key.pem",
            certs_out_uc_dir / "device_priv_key.pem",
        )
        copyfile(certs_out_dir / "ca.pem", certs_out_uc_dir / "ca.pem")

        copyfile(
            certs_out_dir / "certs" / "azure-iot-test-only.root.ca.cert.pem",
            certs_out_azure_dir / f"{device_name}-root-ca.pem",
        )
        logger.info("Cert generated for dev: %s", device_name)
    except FileNotFoundError as e:
        logger.error("Failed to copy certificate %s", e.filename)
        sys.exit(1)


def generate_fs(config: Config) -> None:
    target = config.project.target
    if target.family == "esp32":
        gen_fs_esp32(config)
    else:
        logger.warning("Gen --fs command not supported for this target yet.")


def gen_single_binary(project_profile_path: Path, config: Config) -> None:
    target = config.project.target
    if target.family == "esp32":
        gen_single_binary_esp32(project_profile_path, config)
    else:
        logger.warning("Gen --bin command not supported for this target yet.")


def gen_fs_esp32(config: Config) -> None:
    command = []
    partition_list = config.get_esp32_config().partitions
    fs_size = 0
    for i in partition_list:
        split_list = i.split(",")
        if "spiffs" in split_list[0]:
            fs_size = int(re.sub(",", "", split_list[4]), 16)

    try:
        project_path = config.project_root
        fs_in_dir = project_path / "main/fs"
        fs_out_dir = project_path / OUT_FS_DIR
        fs_out_bin = project_path / "build/spiffs.bin"
        fs_in_dir.mkdir(parents=True, exist_ok=True)
        fs_out_dir.mkdir(parents=True, exist_ok=True)

        shutil.copytree(fs_in_dir, fs_out_dir, dirs_exist_ok=True)

        idf_path = os.environ.get("IDF_PATH")
        command = [
            f"{idf_path}/components/spiffs/spiffsgen.py",
            str(fs_size),
            str(fs_out_dir),
            str(fs_out_bin),
        ]

        subprocess.check_call(command, cwd=project_path)
        logger.info("Generated %s of size:%s", fs_out_bin, fs_size)

    except subprocess.CalledProcessError:
        logger.error("%s fail", command)
        sys.exit(1)


def gen_single_binary_esp32(project_profile_path: Path, config: Config) -> None:
    partition_list = config.get_esp32_config().partitions
    target = config.project.target

    flasher_args_path = project_profile_path / "flash_args"
    if not flasher_args_path.is_file():
        logger.warning("%s does not exists", flasher_args_path)
        sys.exit(1)

    line_list = flasher_args_path.read_text().split()

    spiffs_addr = None
    for i in partition_list:
        if "spiffs" in i:
            split_list = i.split(",")
            spiffs_addr = split_list[3].strip()
            break

    flash_size = "4MB"
    for index, arg in enumerate(line_list):
        if "flash_size" in arg:
            flash_size = line_list[index + 1]
        if arg.endswith(".bin"):
            line_list[index] = str(project_profile_path / arg)

    command = [
        "esptool.py",
        "--chip",
        target.id,
        "merge_bin",
        "-o",
        "build/flash_image.bin",
        "--fill-flash-size",
        f"{flash_size}",
    ]
    command.extend(line_list)

    if spiffs_addr:
        command.extend([spiffs_addr, "build/spiffs.bin"])

    try:
        logger.info("Running: %s", " ".join(command))
        subprocess.check_call(command, cwd=config.project_root)
    except subprocess.CalledProcessError:
        logger.error("Generation of single binary failed")
        sys.exit(1)



# #
# @copyright Copyright (C) 2023 SpyroSoft Solutions S.A. All rights reserved.
# #

import subprocess
import sys
from typing import Optional

from scargo.config import Config
from scargo.config_utils import prepare_config
from scargo.logger import get_logger

logger = get_logger()


def scargo_flash(
    app: bool,
    fs: bool,
    flash_profile: str,
    port: Optional[str] = None,
    erase_memory: bool = True,
) -> None:
    config = prepare_config()
    target = config.project.target

    if port and (target.family != "esp32" and target.family != "stm32"):
        logger.error("--port option is only supported for esp32 and stm32 projects.")
        sys.exit(1)
    if not erase_memory and target.family != "stm32":
        logger.error("--no-erase option is only supported for stm32 projects.")
        sys.exit(1)
    if target.family == "esp32":
        flash_esp32(config, app=app, fs=fs, flash_profile=flash_profile, port=port)
    elif target.family == "stm32":
        flash_stm32(config, flash_profile, erase_memory, port=port)
    else:
        logger.error("Flash command not supported for this target yet.")


def flash_esp32(
    config: Config,
    app: bool,
    fs: bool,
    flash_profile: str = "Debug",
    port: Optional[str] = None,
) -> None:
    project_path = config.project_root
    out_dir = project_path / "build" / flash_profile
    target = config.project.target
    command = []
    try:
        if app:
            app_name = config.project.name
            app_path = out_dir / f"{app_name}.bin"
            command = ["parttool.py"]
            if port:
                command.append(f"--port={port}")
            command.extend(
                [
                    "write_partition",
                    "--partition-name=ota_0",
                    f"--input={app_path}",
                ]
            )
            subprocess.check_call(command)
        elif fs:
            fs_path = config.project_root / "build" / "spiffs.bin"
            command = ["parttool.py"]
            if port:
                command.append(f"--port={port}")
            command.extend(
                [
                    "write_partition",
                    "--partition-name=spiffs",
                    f"--input={fs_path}",
                ]
            )
            subprocess.check_call(command)
        else:
            command = ["esptool.py"]
            if port:
                command.append(f"--port={port}")
            command.extend(["--chip", target.id, "write_flash", "@flash_args"])
            subprocess.check_call(command, cwd=out_dir)
    except subprocess.CalledProcessError:
        logger.error("%s fail", command)


def flash_stm32(
    config: Config,
    flash_profile: str = "Debug",
    erase_memory: bool = True,
    port: Optional[str] = None,
) -> None:
    project_path = config.project_root
    bin_path = (
        project_path
        / "build"
        / flash_profile
        / "bin"
        / f"{config.project.name.lower()}.bin"
    )

    flash_start = hex(config.get_stm32_config().flash_start)

    if not bin_path.exists():
        logger.error("%s does not exist", bin_path)
        logger.info("Did you run scargo build --profile %s", flash_profile)
    elif not flash_start:
        logger.error("Flash start address not found in lock file")
        logger.info("Define flash-start in scargo.toml under stm32 section")
    else:
        if erase_memory:
            command = ["st-flash"]
            if port:
                command.append(f"--serial={port}")
            command.append("erase")
            subprocess.check_call(command)

        command = ["st-flash", "--reset"]
        if port:
            command.append(f"--serial={port}")
        command.extend(["write", str(bin_path), flash_start])
        subprocess.check_call(command)



# #
# @copyright Copyright (C) 2023 SpyroSoft Solutions S.A. All rights reserved.
# #

"""feature function for scargo"""
from scargo import __version__


def scargo_version() -> None:
    print(f"scargo version: {__version__}")



from pathlib import Path

from clang.cindex import Index

from scargo.file_generators.clang_parser.data_classes import HeaderDescriptor
from scargo.file_generators.clang_parser.params_extractor import (
    extract_classes,
    extract_includes,
    extract_namespaces,
)


def parse_file(file_path: Path) -> HeaderDescriptor:
    """TODO documentation
    look here for available types which can be extracted
    https://clang.llvm.org/doxygen/group__CINDEX.html
    This function receives a path with file name as an argument and returns
    list of includes, i.e. every string that starts with "#include",
    dict of classes where keys are "public", "private" and "mock_class_name"
    and list of namespaces, i.e. strings that start with "namespace".
    """
    idx = Index.create()
    translation_unit = idx.parse(str(file_path), ["-x", "c++"])

    namespaces = extract_namespaces(translation_unit.cursor, str(file_path))
    classes = extract_classes(translation_unit.cursor, str(file_path))
    includes = extract_includes(translation_unit.cursor, str(file_path))
    return HeaderDescriptor(
        name=file_path.name, namespaces=namespaces, classes=classes, includes=includes
    )



# #
# @copyright Copyright (C) 2023 SpyroSoft Solutions S.A. All rights reserved.
# #
from typing import Any, Dict

from scargo.config import Config
from scargo.file_generators.base_gen import create_file_from_template


class _CppTemplateGen:
    """
    This class is a container cpp files creation with multilayer approach
    """

    def __init__(self, config: Config) -> None:
        self._config = config
        self._src_dir = config.source_dir_path
        self._inc_dir = config.include_dir_path

    def _generate_bin(self, bin_name: str) -> None:
        """Function which creates main.cpp file using jinja"""
        self._create_file_from_template(
            "cpp/main.cpp.j2",
            f"{bin_name.lower()}.cpp",
            template_params={
                "target": self._config.project.target,
                "bin_name": bin_name,
            },
        )

    def _generate_lib(self, lib_name: str) -> None:
        """Function which creates a lib files using jinja"""

        lib_name = lib_name.lower()
        class_name = "".join(lib_name.replace("_", " ").title().split())

        create_file_from_template(
            "cpp/lib.cpp.j2",
            self._src_dir / f"{lib_name}.cpp",
            template_params={"class_name": class_name, "lib_name": lib_name},
            config=self._config,
        )
        create_file_from_template(
            "cpp/lib.h.j2",
            self._inc_dir / f"{lib_name}.h",
            template_params={"class_name": class_name},
            config=self._config,
        )

    def _generate_cmake(self) -> None:
        self._create_file_from_template(
            f"cpp/cmake-src-{self._config.project.target.family}.j2",
            "CMakeLists.txt",
            template_params={"config": self._config},
        )

    def _create_file_from_template(
        self,
        template_path: str,
        output_filename: str,
        template_params: Dict[str, Any],
    ) -> None:
        create_file_from_template(
            template_path,
            self._src_dir / output_filename,
            template_params=template_params,
            config=self._config,
        )

    def generate_cpp(self) -> None:
        """Generate dirs and files"""
        lib_name = self._config.project.lib_name
        if lib_name:
            self._generate_lib(lib_name)

        bin_name = self._config.project.bin_name
        if bin_name:
            self._generate_bin(bin_name)

        self._generate_cmake()


def generate_cpp(config: Config) -> None:
    cpp_template = _CppTemplateGen(config)
    cpp_template.generate_cpp()



# #
# @copyright Copyright (C) 2023 SpyroSoft Solutions S.A. All rights reserved.
# #
from enum import Enum
from pathlib import Path
from typing import Any, Dict, List, Optional

import toml
from pydantic import BaseModel, Extra, Field, root_validator

from scargo.global_values import SCARGO_DEFAULT_BUILD_ENV


class ConfigError(Exception):
    pass


class Config(BaseModel):
    project: "ProjectConfig"
    profiles: Dict[str, "ProfileConfig"] = Field(..., alias="profile")
    check: "ChecksConfig"
    doc: "DocConfig" = Field(
        default_factory=lambda: DocConfig()  # pylint: disable=unnecessary-lambda
    )
    tests: "TestConfig"
    dependencies: "Dependencies"
    conan: "ConanConfig"
    stm32: Optional["Stm32Config"]
    esp32: Optional["Esp32Config"]
    scargo: "ScargoConfig" = Field(
        default_factory=lambda: ScargoConfig()  # pylint: disable=unnecessary-lambda
    )
    docker_compose: "DockerComposeConfig" = Field(
        default_factory=lambda: DockerComposeConfig(),  # pylint: disable=unnecessary-lambda
        alias="docker-compose",
    )
    project_root: Path

    @property
    def source_dir_path(self) -> Path:
        return self.project_root / self.project.target.source_dir

    @property
    def include_dir_path(self) -> Path:
        return self.project_root / self.project.target.include_dir

    def get_stm32_config(self) -> "Stm32Config":
        if not self.stm32:
            raise ConfigError("No [stm32] section in config")
        return self.stm32

    def get_esp32_config(self) -> "Esp32Config":
        if not self.esp32:
            raise ConfigError("No [esp32] section in config")
        return self.esp32

    @root_validator
    def validate_special_configs(  # pylint: disable=no-self-argument
        cls, values: Dict[str, Any]
    ) -> Dict[str, Any]:
        if "project" in values:
            target_id = values["project"].target_id
            if target_id == "stm32" and not values["stm32"]:
                raise ConfigError("No [stm32] section in config")
            if target_id == "esp32" and not values["esp32"]:
                raise ConfigError("No [esp32] section in config")
        return values


class ProjectConfig(BaseModel):
    name: str
    version: str
    description: Optional[str]
    homepage_url: Optional[str] = Field(None, alias="homepage-url")

    bin_name: Optional[str]
    lib_name: Optional[str]
    target_id: str = Field(..., alias="target")
    build_env: str = Field(SCARGO_DEFAULT_BUILD_ENV, alias="build-env")
    docker_file: Path = Field(..., alias="docker-file")
    docker_image_tag: str = Field(..., alias="docker-image-tag")
    in_repo_conan_cache: bool = Field(..., alias="in-repo-conan-cache")

    cc: Optional[str] = None
    cxx: Optional[str] = None
    cxxstandard: str

    cflags: str
    cxxflags: str

    max_build_jobs: Optional[int] = Field(None, alias="max-build-jobs")

    cmake_variables: Dict[str, str] = Field(
        default_factory=dict, alias="cmake-variables"
    )

    @property
    def target(self) -> "Target":
        return Target.get_target_by_id(self.target_id)

    def get_compiler_warning(self) -> Optional[str]:
        if (self.cc and not self.target.cc) or (self.cxx and not self.target.cxx):
            return "Compiler settings are ignored for this target"
        return None


class Target(BaseModel):
    id: str
    family: str
    source_dir: str
    include_dir: str
    cc: Optional[str] = None
    cxx: Optional[str] = None

    @classmethod
    def get_target_by_id(cls, target_id: str) -> "Target":
        return TARGETS[target_id]


TARGETS = {
    "x86": Target(
        id="x86",
        family="x86",
        source_dir="src",
        include_dir="include",
        cc="gcc",
        cxx="g++",
    ),
    "stm32": Target(
        id="stm32", family="stm32", source_dir="src", include_dir="include"
    ),
    "esp32": Target(
        id="esp32", family="esp32", source_dir="main", include_dir="include"
    ),
    "esp32s2": Target(
        id="esp32s2", family="esp32", source_dir="main", include_dir="include"
    ),
    "esp32s3": Target(
        id="esp32s3", family="esp32", source_dir="main", include_dir="include"
    ),
    "esp32c2": Target(
        id="esp32c2", family="esp32", source_dir="main", include_dir="include"
    ),
    "esp32c3": Target(
        id="esp32c3", family="esp32", source_dir="main", include_dir="include"
    ),
}


# for typer
ScargoTargets = Enum(  # type: ignore[misc]
    "ScargoTargets", {target.id: target.id for target in TARGETS.values()}
)


class ProfileConfig(BaseModel, extra=Extra.allow):
    cflags: Optional[str]
    cxxflags: Optional[str]
    cc: Optional[str] = None
    cxx: Optional[str] = None
    cmake_build_type: Optional[str] = Field("Debug", alias="cmake-build-type")

    @property
    def extras(self) -> Dict[str, Any]:
        return {
            key: value
            for key, value in dict(self).items()
            if key not in self.__fields__
        }


class ChecksConfig(BaseModel):
    exclude: List[str]
    pragma: "CheckConfig"
    copyright: "CheckConfig"
    todo: "TodoCheckConfig"
    clang_format: "CheckConfig" = Field(..., alias="clang-format")
    clang_tidy: "CheckConfig" = Field(..., alias="clang-tidy")
    cyclomatic: "CheckConfig"


class CheckConfig(BaseModel):
    description: Optional[str]
    exclude: List[str] = Field(default_factory=list)


class TodoCheckConfig(CheckConfig):
    keywords: List[str] = Field(default_factory=list)


class DocConfig(BaseModel):
    exclude: List[str] = Field(default_factory=list)


class TestConfig(BaseModel, extra=Extra.allow):
    cc: str
    cxx: str

    cflags: str
    cxxflags: str

    gcov_executable: str = Field(..., alias="gcov-executable")

    @property
    def extras(self) -> Dict[str, Any]:
        return {
            key: value
            for key, value in dict(self).items()
            if key not in self.__fields__
        }


class Dependencies(BaseModel):
    general: List[str] = Field(default_factory=list)
    build: List[str] = Field(default_factory=list)
    tool: List[str] = Field(default_factory=list)
    test: List[str] = Field(default_factory=list)


class ConanConfig(BaseModel):
    repo: Dict[str, str]  # name to url mapping


class Stm32Config(BaseModel):
    chip: str
    flash_start: int = Field(alias="flash-start")


class Esp32Config(BaseModel):
    extra_component_dirs: List[Path] = Field(default_factory=list)
    partitions: List[str] = Field(default_factory=list)


class ScargoConfig(BaseModel):
    console_log_level: str = Field("INFO", alias="console-log-level")
    file_log_level: str = Field("WARNING", alias="file-log-level")
    update_exclude: List[str] = Field(alias="update-exclude", default_factory=list)
    version: Optional[str]


class DockerComposeConfig(BaseModel):
    ports: List[str] = Field(default_factory=list)


Config.update_forward_refs()
ProjectConfig.update_forward_refs()
ChecksConfig.update_forward_refs()
Stm32Config.update_forward_refs()
Esp32Config.update_forward_refs()


def parse_config(config_file_path: Path) -> Config:
    config_obj = toml.load(config_file_path)
    config_obj["project_root"] = config_file_path.parent.absolute()
    return Config.parse_obj(config_obj)



from typing import List

from clang.cindex import Cursor, CursorKind

from scargo.file_generators.clang_parser.data_classes import (
    ArgumentDescriptor,
    ClassDescriptor,
    FunctionDescriptor,
    IncludeDescriptor,
    NamespaceDescriptor,
)


def extract_namespaces(cursor: Cursor, filename: str) -> List[NamespaceDescriptor]:
    return [
        NamespaceDescriptor(descendant.spelling)
        for descendant in cursor.walk_preorder()
        if descendant.kind == CursorKind.NAMESPACE
        # remove type ignore after fix is relased:
        # https://github.com/tgockel/types-clang/issues/10
        and descendant.location.file.name == filename  # type: ignore[attr-defined]
    ]


def extract_classes(cursor: Cursor, filename: str) -> List[ClassDescriptor]:
    return [
        ClassDescriptor(
            descendant.spelling,
            f"Mock{descendant.spelling}",
            _extract_cxx_methods(descendant),
        )
        for descendant in cursor.walk_preorder()
        if descendant.kind == CursorKind.CLASS_DECL
        and descendant.location.file.name == filename  # type: ignore[attr-defined]
    ]


def extract_includes(cursor: Cursor, filename: str) -> List[IncludeDescriptor]:
    return [
        IncludeDescriptor(descendant.displayname)
        for descendant in cursor.walk_preorder()
        if descendant.kind == CursorKind.INCLUSION_DIRECTIVE
        and descendant.location.file.name == filename  # type: ignore[attr-defined]
    ]


def _extract_cxx_methods(cursor: Cursor) -> List[FunctionDescriptor]:
    return [
        _extract_method_params(descendant)
        for descendant in cursor.walk_preorder()
        if descendant.kind == CursorKind.CXX_METHOD
        and descendant.access_specifier.name == "PUBLIC"
    ]


def _extract_method_params(
    cursor: Cursor,
) -> FunctionDescriptor:
    specifiers = []
    if cursor.is_virtual_method():
        specifiers.append("override")
    if cursor.is_const_method():
        specifiers.append("const")

    fun_args = [
        ArgumentDescriptor(descendant.spelling, descendant.type.spelling)
        for descendant in cursor.walk_preorder()
        if descendant.kind == CursorKind.PARM_DECL
    ]
    return FunctionDescriptor(
        name=cursor.spelling,
        return_type=cursor.result_type.spelling,
        specifiers=specifiers,
        arguments=fun_args,
    )



# #
# @copyright Copyright (C) 2023 SpyroSoft Solutions S.A. All rights reserved.
# #

"""Create project documentation"""
import subprocess
import sys
from pathlib import Path
from typing import Iterable

import typer

from scargo.config import Config
from scargo.config_utils import prepare_config
from scargo.logger import get_logger
from scargo.path_utils import find_program_path

logger = get_logger()


class _ScargoGenDoc:
    EXCLUDE_LIST = ["build"]

    def __init__(self, config: Config, doc_dir_path: Path):
        self._config = config
        self._doc_dir_path = doc_dir_path

    def create_default_doxyfile(self) -> None:
        """Create default doxyfile"""
        subprocess.check_call("doxygen -g", shell=True, cwd=self._doc_dir_path)

    def update_doxyfile(self) -> None:
        """Update Doxyfile configuration according specified values"""
        doxyfile_path = self._doc_dir_path / "Doxyfile"
        with doxyfile_path.open(encoding="utf-8") as doxyfile:
            doxyfile_lines = doxyfile.readlines()

        with doxyfile_path.open("w", encoding="utf-8") as doxyfile:
            doxyfile.writelines(self.adjust_doxyfile_lines(doxyfile_lines))

    def adjust_doxyfile_lines(self, lines: Iterable[str]) -> Iterable[str]:
        project_name = self._config.project.name
        project_path = self._config.project_root
        exclude = " ".join(
            f"{project_path}/{dir}"
            for dir in self.EXCLUDE_LIST + self._config.doc.exclude
        )
        doxy_values = {
            "PROJECT_NAME": f'"{project_name}"',
            "EXTRACT_ALL": "YES",
            "INPUT": project_path,
            "RECURSIVE": "YES",
            "EXCLUDE_PATTERNS": exclude,
            "GENERATE_LATEX": "NO",
        }
        for line in lines:
            key = line.split(" ", 1)[0]
            if key in doxy_values:
                yield f"{key.ljust(23)}= {doxy_values[key]}\n"
            yield line

    def generate_doxygen(self) -> None:
        """Generate doxygen according to doxyfile"""
        subprocess.check_call("doxygen", shell=True, cwd=self._doc_dir_path)


def _open_doc(doc_dir_path: Path) -> None:
    html_file_path = doc_dir_path / "html/index.html"
    if html_file_path.exists():
        try:
            typer.launch(str(html_file_path))
        except subprocess.CalledProcessError:
            logger.error("Fail to open documentation")
    else:
        logger.error("Documentation not found")
        logger.info("Create documentation using command: scargo doc")
        sys.exit(1)


def scargo_doc(open_doc: bool) -> None:
    """
    Create documentation for project

    :param bool open_doc: if true open documentation instead creating it
    :return: None
    """
    config = prepare_config()
    doc_dir_path = config.project_root / "build/doc"
    if open_doc:
        _open_doc(doc_dir_path)
        sys.exit(0)

    if not find_program_path("doxygen"):
        logger.error("Doxygen not installed or not in PATH environment variable")
        sys.exit(1)

    doc_dir_path.mkdir(parents=True, exist_ok=True)

    try:
        scargo_doc_gen = _ScargoGenDoc(config, doc_dir_path)
        scargo_doc_gen.create_default_doxyfile()
        scargo_doc_gen.update_doxyfile()
        scargo_doc_gen.generate_doxygen()

    except subprocess.CalledProcessError:
        logger.error("Fail to create project documentation")



# #
# @copyright Copyright (C) 2023 SpyroSoft Solutions S.A. All rights reserved.
# #
import os
import sys
from pathlib import Path
from typing import List, Optional

from typer import Argument, Option, Typer

from scargo.commands.build import scargo_build
from scargo.commands.check import scargo_check
from scargo.commands.clean import scargo_clean
from scargo.commands.debug import scargo_debug
from scargo.commands.doc import scargo_doc
from scargo.commands.docker import (
    scargo_docker_build,
    scargo_docker_exec,
    scargo_docker_run,
)
from scargo.commands.fix import scargo_fix
from scargo.commands.flash import scargo_flash
from scargo.commands.gen import scargo_gen
from scargo.commands.new import scargo_new
from scargo.commands.publish import scargo_publish
from scargo.commands.run import scargo_run
from scargo.commands.test import scargo_test
from scargo.commands.update import scargo_update
from scargo.commands.version import scargo_version
from scargo.config import ScargoTargets, Target
from scargo.global_values import DESCRIPTION, SCARGO_DEFAULT_CONFIG_FILE
from scargo.logger import get_logger
from scargo.path_utils import get_config_file_path

logger = get_logger()

###############################################################################


cli = Typer(context_settings={"help_option_names": ["-h", "--help"]}, help=DESCRIPTION)


BASE_DIR_OPTION = Option(
    None,
    "--base-dir",
    "-B",
    exists=True,
    file_okay=False,
    help="Base directory of the project",
)


###############################################################################


@cli.command()
def build(
    profile: str = Option("Debug", "--profile"),
    base_dir: Optional[Path] = BASE_DIR_OPTION,
) -> None:
    """Compile sources."""
    if base_dir:
        os.chdir(base_dir)
    scargo_build(profile)


###############################################################################


@cli.command()
def check(
    clang_format: bool = Option(False, "--clang-format", help="Run clang-format."),
    clang_tidy: bool = Option(False, "--clang-tidy", help="Run clang-tidy."),
    copy_right: bool = Option(False, "--copyright", help="Run copyright check."),
    cppcheck: bool = Option(False, "--cppcheck", help="Run cppcheck."),
    cyclomatic: bool = Option(False, "--cyclomatic", help="Run python-lizard."),
    pragma: bool = Option(False, "--pragma", help="Run pragma check."),
    todo: bool = Option(False, "--todo", help="Run TODO check."),
    silent: bool = Option(False, "--silent", "-s", help="Show less output."),
    base_dir: Optional[Path] = BASE_DIR_OPTION,
) -> None:
    """Check source code in directory `src`."""
    if base_dir:
        os.chdir(base_dir)
    scargo_check(
        clang_format,
        clang_tidy,
        copy_right,
        cppcheck,
        cyclomatic,
        pragma,
        todo,
        verbose=not silent,
    )


###############################################################################


@cli.command()
def clean(base_dir: Optional[Path] = BASE_DIR_OPTION) -> None:
    """Remove directory `build`."""
    if base_dir:
        os.chdir(base_dir)
    scargo_clean()


###############################################################################


@cli.command()
def debug(
    bin_path: Optional[Path] = Option(
        None,
        "--bin",
        "-b",
        exists=True,
        dir_okay=False,
        resolve_path=True,
        help="Path to bin file",
    ),
    base_dir: Optional[Path] = BASE_DIR_OPTION,
) -> None:
    """Use gdb cli to debug"""
    if base_dir:
        os.chdir(base_dir)
    scargo_debug(bin_path)


###############################################################################


@cli.command()
def doc(
    open_doc: bool = Option(False, "--open", help="Open html documentation"),
    base_dir: Optional[Path] = BASE_DIR_OPTION,
) -> None:
    """Create project documentation"""
    if base_dir:
        os.chdir(base_dir)
    scargo_doc(open_doc)


###############################################################################


docker = Typer(help="Manage the docker environment for the project")


@docker.command(
    "build", context_settings={"allow_extra_args": True, "ignore_unknown_options": True}
)
def docker_build(
    docker_opts: List[str] = Argument(None), base_dir: Optional[Path] = BASE_DIR_OPTION
) -> None:
    """Build docker layers for this project depending on the target"""
    if base_dir:
        os.chdir(base_dir)
    scargo_docker_build(docker_opts)


@docker.command(
    "run", context_settings={"allow_extra_args": True, "ignore_unknown_options": True}
)
def docker_run(
    command: str = Option(
        "bash",
        "-c",
        "--command",
        metavar="COMMAND",
        help="Select command to be used with docker run.",
    ),
    base_dir: Optional[Path] = BASE_DIR_OPTION,
    docker_opts: List[str] = Argument(None),
) -> None:
    """Run project in docker environment"""
    if base_dir:
        os.chdir(base_dir)
    scargo_docker_run(docker_opts=docker_opts, command=command)


@docker.command(
    "exec", context_settings={"allow_extra_args": True, "ignore_unknown_options": True}
)
def docker_exec(
    base_dir: Optional[Path] = BASE_DIR_OPTION, docker_opts: List[str] = Argument(None)
) -> None:
    """Attach to existing docker environment"""
    if base_dir:
        os.chdir(base_dir)
    scargo_docker_exec(docker_opts)


cli.add_typer(docker, name="docker")


###############################################################################


@cli.command()
def fix(
    clang_format: bool = Option(
        False, "--clang-format", help="Fix clang-format violations"
    ),
    copy_right: bool = Option(False, "--copyright", help="Fix copyrights violations"),
    pragma: bool = Option(False, "--pragma", help="Fix pragma violations"),
    base_dir: Optional[Path] = BASE_DIR_OPTION,
) -> None:
    """Fix violations reported by command `check`."""
    if base_dir:
        os.chdir(base_dir)
    scargo_fix(pragma, copy_right, clang_format)


###############################################################################


@cli.command()
def flash(
    app: bool = Option(False, "--app", help="Flash app only"),
    file_system: bool = Option(False, "--fs", help="Flash filesystem only"),
    flash_profile: str = Option(
        "Debug", "--profile", help="Flash base on previously built profile"
    ),
    port: Optional[str] = Option(
        None,
        help="(esp32 only) port where the target device of the command is connected to, e.g. /dev/ttyUSB0",
    ),
    no_erase: bool = Option(False, help="(stm32 only) Don't erase target memory"),
    base_dir: Optional[Path] = BASE_DIR_OPTION,
) -> None:
    """Flash the target."""
    if base_dir:
        os.chdir(base_dir)
    scargo_flash(app, file_system, flash_profile, port, not no_erase)


###############################################################################


@cli.command()
def gen(
    profile: str = Option("Debug", "--profile", "-p"),
    gen_ut: Optional[Path] = Option(
        None,
        "--unit-test",
        "-u",
        exists=True,
        resolve_path=True,
        help="Generate unit test for chosen file or all headers in directory",
    ),
    gen_mock: Optional[Path] = Option(
        None,
        "--mock",
        "-m",
        exists=True,
        dir_okay=False,
        resolve_path=True,
        help="Generate mock of chosen file",
    ),
    certs: Optional[str] = Option(
        None,
        "--certs",
        "-c",
        metavar="<DEVICE ID>",
        help="Generate cert files for azure based on device ID.",
    ),
    certs_mode: Optional[str] = Option(
        None,
        "--type",
        "-t",
        metavar="[all, device]",
        help="Mode for generating certificates.",
    ),
    certs_input: Optional[Path] = Option(
        None,
        "--in",
        "-i",
        dir_okay=True,
        resolve_path=True,
        help="Directory with root and intermediate certificates.",
    ),
    certs_passwd: Optional[str] = Option(
        None,
        "--passwd",
        "-p",
        metavar="<PASSWORD>",
        help="Password to be set for generated certificates",
    ),
    file_system: bool = Option(
        False, "--fs", "-f", help="Build the filesystem, base on spiffs dir content."
    ),
    single_bin: bool = Option(
        False, "--bin", "-b", help="Generate single binary image."
    ),
    base_dir: Optional[Path] = BASE_DIR_OPTION,
) -> None:
    """Manage the auto file generator"""
    if base_dir:
        os.chdir(base_dir)
    if (gen_ut is gen_mock is certs is None) and not (file_system or single_bin):
        logger.warning(
            "Please add one of the following options to the command:"
            "\n--unit-test\n--mock\n--certs\n--fs\n--bin"
        )
        sys.exit(1)

    scargo_gen(
        profile,
        gen_ut,
        gen_mock,
        certs,
        certs_mode,
        certs_input,
        certs_passwd,
        file_system,
        single_bin,
    )


###############################################################################


@cli.command()
def new(
    name: str,
    bin_name: Optional[str] = Option(
        None,
        "--bin",
        help="Create binary target template.",
        prompt=True,
        prompt_required=False,
    ),
    lib_name: Optional[str] = Option(
        None,
        "--lib",
        help="Create library target template.",
        prompt=True,
        prompt_required=False,
    ),
    target: ScargoTargets = Option("x86", help="Target device."),
    create_docker: bool = Option(
        True, "--docker/--no-docker", help="Initialize docker environment."
    ),
    git: bool = Option(True, "--git/--no-git", help="Initialize git repository."),
    base_dir: Optional[Path] = BASE_DIR_OPTION,
) -> None:
    """Create new project template."""
    if base_dir:
        os.chdir(base_dir)
    scargo_new(
        name,
        bin_name,
        lib_name,
        Target.get_target_by_id(target.value),
        create_docker,
        git,
    )
    scargo_update(Path(name, SCARGO_DEFAULT_CONFIG_FILE).absolute())


###############################################################################


@cli.command()
def publish(
    repo: str = Option("", "--repo", "-r", help="Repo name"),
    profile: str = Option("Release", "--profile", "-p"),
    base_dir: Optional[Path] = BASE_DIR_OPTION,
) -> None:
    """Upload conan pkg to repo"""
    if base_dir:
        os.chdir(base_dir)
    scargo_publish(repo, profile)


###############################################################################


@cli.command()
def run(
    bin_path: Optional[Path] = Option(
        None,
        "--bin",
        "-b",
        exists=True,
        dir_okay=False,
        resolve_path=True,
        help="Path to bin file",
    ),
    profile: str = Option("Debug", "--profile", "-p"),
    skip_build: bool = Option(False, "--skip-build", help="Skip calling scargo build"),
    bin_params: List[str] = Argument(None),
    base_dir: Optional[Path] = BASE_DIR_OPTION,
) -> None:
    """Build and run project"""
    if base_dir:
        os.chdir(base_dir)
    if not skip_build:
        scargo_build(profile)
    scargo_run(bin_path, profile, bin_params)


###############################################################################


@cli.command()
def test(
    verbose: bool = Option(False, "--verbose", "-v", help="Verbose mode."),
    profile: str = Option("Debug", "--profile", help="CMake profile to use"),
    detailed_coverage: bool = Option(
        False, help="Generate detailed coverage HTML files"
    ),
    base_dir: Optional[Path] = BASE_DIR_OPTION,
) -> None:
    """Compile and run all tests in directory `test`."""
    if base_dir:
        os.chdir(base_dir)
    scargo_test(verbose, profile, detailed_coverage)


###############################################################################


@cli.command()
def update(
    config_file_path: Optional[Path] = Option(
        None,
        "--config-file",
        "-c",
        exists=True,
        dir_okay=False,
        resolve_path=True,
        help="Path to .toml configuration file.",
    ),
    base_dir: Optional[Path] = BASE_DIR_OPTION,
) -> None:
    """Read .toml config file and generate `CMakeLists.txt`."""
    if base_dir:
        os.chdir(base_dir)
    if config_file_path is None:
        config_file_path = get_config_file_path(SCARGO_DEFAULT_CONFIG_FILE)
        if not config_file_path:
            logger.error("Config file not found.")
            sys.exit(1)
    scargo_update(config_file_path)


###############################################################################


@cli.command()
def version() -> None:
    """Get scargo version"""
    scargo_version()


###############################################################################

if __name__ == "__main__":
    cli()



# #
# @copyright Copyright (C) 2023 SpyroSoft Solutions S.A. All rights reserved.
# #


"""Publish conan package into repository"""
import os
import subprocess
import sys
from pathlib import Path

from scargo.config import Config
from scargo.config_utils import prepare_config
from scargo.logger import get_logger

logger = get_logger()


def scargo_publish(repo: str, profile: str = "Release") -> None:
    """
    Publish conan package

    :param str repo: repository name
    :return: None
    """
    config = prepare_config()
    project_path = config.project_root
    project_config = config.project
    project_name = project_config.name

    conan_clean_remote()
    conan_add_remote(project_path, config)
    conan_add_conancenter()

    # Export package
    try:
        subprocess.check_call(
            [
                "conan",
                "create",
                ".",
                "-pr:b",
                "default",
                "-pr:h",
                f"./.conan/profiles/{config.project.target.family}_{profile}",
                "-b",
                "missing",
            ],
            cwd=project_path,
        )
    except subprocess.CalledProcessError:
        logger.error("Unable to create package")
        sys.exit(1)

    # Upload package to artifactory
    conan_repo = ["-r", repo] if repo else []
    try:
        subprocess.check_call(
            [
                "conan",
                "upload",
                f"{project_name}",
                *conan_repo,
                "--all",
                "--confirm",
            ],
            cwd=project_path,
        )
    except subprocess.CalledProcessError:
        logger.error("Unable to publish package")
        sys.exit(1)


def conan_add_remote(project_path: Path, config: Config) -> None:
    """
    Add conan remote repository

    :param Path project_path: path to project
    :param Config config:
    :return: None
    """
    conan_repo = config.conan.repo
    for repo_name, repo_url in conan_repo.items():
        try:
            subprocess.check_call(
                ["conan", "remote", "add", repo_name, repo_url],
                cwd=project_path,
            )
        except subprocess.CalledProcessError:
            logger.error("Unable to add remote repository")
        conan_add_user(repo_name)


def conan_add_conancenter() -> None:
    """
    Add conancenter remote repository

    :return: None
    """
    try:
        subprocess.check_call(
            "conan remote add conancenter https://center.conan.io", shell=True
        )
    except subprocess.CalledProcessError:
        logger.error("Unable to add conancenter remote repository")


def conan_clean_remote() -> None:
    """
    Clean all remote repositories

    :return: None
    """
    try:
        subprocess.check_call("conan remote clean", shell=True)
    except subprocess.CalledProcessError:
        logger.error("Unable to clean remote repository list")


def conan_add_user(remote: str) -> None:
    """
    Add conan user

    :param str remote: name of remote repository
    :return: None
    """
    conan_user = subprocess.run(
        "conan user", capture_output=True, shell=True, check=False
    ).stdout.decode("utf-8")

    env_conan_user = os.environ.get("CONAN_LOGIN_USERNAME", "")
    env_conan_passwd = os.environ.get("CONAN_PASSWORD", "")

    if env_conan_user not in conan_user:
        try:
            subprocess.check_call(
                ["conan", "user", "-p", env_conan_passwd, "-r", remote, env_conan_user],
            )
        except subprocess.CalledProcessError:
            logger.error("Unable to add user")



# #
# @copyright Copyright (C) 2023 SpyroSoft Solutions S.A. All rights reserved.
# #

import getpass
import os
from dataclasses import dataclass
from pathlib import Path

from scargo.config import Config
from scargo.file_generators.base_gen import create_file_from_template
from scargo.global_values import ENV_DEFAULT_NAME


@dataclass
class _EnvironmentDescriptor:
    user_name: str = getpass.getuser()
    user_passwd: str = "user"
    user_gid: int = 1000 if os.name == "nt" else os.getgid()
    user_uid: int = 1000 if os.name == "nt" else os.getuid()
    ssh_port: int = 2000
    base_docker_image: str = "ubuntu:20.04"
    conan_username: str = ""
    conan_passwd: str = ""


def generate_env(output_dir_path: Path, config: Config) -> None:
    """
    Generate .env file which is used by docker compose
    providing environmental variables
    """
    env_output_path = output_dir_path / ENV_DEFAULT_NAME
    create_file_from_template(
        "docker/env.txt.j2",
        env_output_path,
        overwrite=False,
        template_params={"env": _EnvironmentDescriptor()},
        config=config,
    )



# #
# @copyright Copyright (C) 2023 SpyroSoft Solutions S.A. All rights reserved.
# #

"""Generate cmake for test dir"""
from pathlib import Path
from shutil import copytree

from scargo.config import Config
from scargo.file_generators.base_gen import create_file_from_template
from scargo.global_values import SCARGO_PKG_PATH


def generate_tests(config: Config) -> None:
    """Generate dirs and files"""
    tests_template_dir = Path(SCARGO_PKG_PATH, "file_generators", "templates", "tests")

    # List of files to generate once (template_path, output_path)
    gen_once_file_list = [
        ("tests/CMakeLists-ut.txt.j2", "tests/ut/CMakeLists.txt"),
        ("tests/CMakeLists-it.txt.j2", "tests/it/CMakeLists.txt"),
        ("tests/CMakeLists-mocks.txt.j2", "tests/mocks/CMakeLists.txt"),
    ]

    static_mock_dir = config.project_root / "tests" / "mocks" / "static_mock"
    if not static_mock_dir.exists():
        copytree(tests_template_dir / "static_mock", static_mock_dir)

    # Update main test cmake on scargo update
    create_file_from_template(
        "tests/CMakeLists-test.txt.j2",
        "tests/CMakeLists.txt",
        overwrite=True,
        template_params={
            "target": config.project.target,
            "tests": config.tests,
            "cxxstandard": config.project.cxxstandard,
        },
        config=config,
    )

    for template, output_path in gen_once_file_list:
        create_file_from_template(
            template,
            output_path,
            overwrite=False,
            template_params={},
            config=config,
        )



# #
# @copyright Copyright (C) 2023 SpyroSoft Solutions S.A. All rights reserved.
# #
from pathlib import Path
from typing import List, Sequence

from scargo.config import Config
from scargo.file_generators.base_gen import create_file_from_template
from scargo.file_generators.clang_parser.header_parser import parse_file

HEADER_EXTENSIONS = (".h", ".hpp")
SRC_EXTENSIONS = (".c", ".cpp")


class _UnitTestsGen:
    def __init__(self, config: Config):
        self._config = config
        self._project_path = config.project_root
        self._ut_dir = self._project_path / "tests/ut"

    def generate_tests(self, input_path: Path, overwrite: bool) -> None:
        """Generates unit test files and corresponding cmake file

        :param Path input_path: Path to src file or src directory
        :param bool overwrite: overwrite files if exist
        """
        if input_path.is_file():
            ut_path = self._get_unit_test_path(input_path)
            self._generate_unit_test(input_path, ut_path, overwrite)
            self._generate_cmake(input_path.parent, ut_path.parent)

        elif input_path.is_dir():
            headers = self._get_paths_with_ext(input_path, HEADER_EXTENSIONS)
            ut_path = None
            for hdr in headers:
                ut_path = self._get_unit_test_path(hdr)
                self._generate_unit_test(hdr, ut_path, overwrite)
            if ut_path:
                self._generate_cmake(input_path, ut_path.parent)

    def _generate_unit_test(
        self, input_file_path: Path, output_file_path: Path, overwrite: bool
    ) -> None:
        """Generates unit test source file

        :param Path input_file_path: Path to src file
        :param Path output_file_path: Path to unit test file
        :param bool overwrite: overwrite if exists
        """
        header_descriptor = parse_file(input_file_path)
        create_file_from_template(
            "ut/ut.cpp.j2",
            output_file_path,
            overwrite=overwrite,
            template_params={"header": header_descriptor},
            config=self._config,
        )

    def _generate_cmake(self, src_dir_path: Path, ut_dir_path: Path) -> None:
        """Generate CMakeLists for unit tests

        :param Path src_dir_path: Source directory for which tests are being generated
        :param Path ut_dir_path: Directory of generated unit tests
        """
        cmake_dir_path = ut_dir_path
        while cmake_dir_path != self._ut_dir:
            add_ut_dir_to_parent_cmake(cmake_dir_path)
            cmake_dir_path = cmake_dir_path.parent

        ut_name = self._get_cmake_tests_name(ut_dir_path)
        ut_files = [
            p.name for p in self._get_paths_with_ext(ut_dir_path, SRC_EXTENSIONS)
        ]

        # Exclude main from srcs to test
        main_cpp = (
            f"{self._config.project.bin_name}.cpp"
            if self._config.project.bin_name
            else None
        )

        src_files = [
            p.relative_to(self._project_path)
            for p in self._get_paths_with_ext(src_dir_path, SRC_EXTENSIONS)
            if p.name != main_cpp
        ]

        create_file_from_template(
            "ut/CMakeLists.txt.j2",
            ut_dir_path / "CMakeLists.txt",
            overwrite=True,
            template_params={
                "src_files": src_files,
                "utest_name": ut_name,
                "ut_files": ut_files,
            },
            config=self._config,
        )

    def _get_unit_test_path(self, input_src_path: Path) -> Path:
        """Return output path for unit test

        :param Path input_src_path: Source path
        :return Path: output path for unit test
        """
        relative_to_src = input_src_path.relative_to(self._config.source_dir_path)
        return Path(self._ut_dir, relative_to_src).with_name(
            f"ut_{input_src_path.stem}.cpp"
        )

    def _get_cmake_tests_name(self, test_dir_path: Path) -> str:
        """Get tests name for cmake

        :param Path test_dir_path: Path to test dir
        :return str: tests name for cmake
        """
        relative_path = test_dir_path.relative_to(self._project_path)
        return "_".join(relative_path.parts)

    @staticmethod
    def _get_paths_with_ext(workdir: Path, extensions: Sequence[str]) -> List[Path]:
        return [child for child in workdir.iterdir() if child.suffix in extensions]


COPYRIGHT = """# #
# Copyright (C) 2022 Spyrosoft Solutions. All rights reserved.
# #\n\n"""


def add_ut_dir_to_parent_cmake(ut_dir_path: Path) -> None:
    # if CMakelists doesn't exist create it in parent dir
    parent_dir = ut_dir_path.parent
    cmake_list_path = parent_dir / "CMakeLists.txt"
    if not cmake_list_path.exists():
        with cmake_list_path.open("w", encoding="utf-8") as file:
            file.write(COPYRIGHT)
    # check if ut_dir_path exists in CMakeLists
    # if not add it
    with cmake_list_path.open("r+", encoding="utf-8") as file:
        cmake_text = file.read()
        add_subdirectory_clause = f"\nadd_subdirectory({ut_dir_path.name})"
        if add_subdirectory_clause not in cmake_text:
            file.write(add_subdirectory_clause)


def generate_ut(input_path: Path, config: Config, force: bool = False) -> None:
    ut_gen = _UnitTestsGen(config)
    ut_gen.generate_tests(input_path, force)



# #
# @copyright Copyright (C) 2023 SpyroSoft Solutions S.A. All rights reserved.
# #

import logging
from typing import Tuple

import coloredlogs

from scargo.config import parse_config
from scargo.global_values import SCARGO_LOCK_FILE
from scargo.path_utils import get_config_file_path, get_project_root_or_none


def __get_logging_config() -> Tuple[int, int]:
    console_log_level = logging.INFO
    file_log_level = logging.WARNING
    try:
        lock_file = get_config_file_path(SCARGO_LOCK_FILE)
        if not lock_file:
            return console_log_level, file_log_level
        config = parse_config(lock_file)
        scargo_config = config.scargo
        console_log_level = logging.getLevelName(scargo_config.console_log_level)
        file_log_level = logging.getLevelName(scargo_config.file_log_level)
    finally:
        return console_log_level, file_log_level  # pylint: disable=lost-exception


def get_logger(name: str = "scargo") -> logging.Logger:
    if logging.getLogger(name).handlers:
        return logging.getLogger(name)

    console_log_level, file_log_level = __get_logging_config()
    logger = logging.getLogger(name)
    logger.setLevel(min(console_log_level, file_log_level))

    stream_handler = logging.StreamHandler()
    stream_handler.addFilter(coloredlogs.HostNameFilter())
    stream_formatter = coloredlogs.ColoredFormatter(
        fmt="%(name)s %(levelname)4s: %(message)s"
    )
    stream_handler.setLevel(console_log_level)
    stream_handler.setFormatter(stream_formatter)
    logger.addHandler(stream_handler)

    project_root = get_project_root_or_none()
    if project_root:
        log_path = project_root / f"{name}.log"
        file_handler = logging.FileHandler(log_path)
        file_handler.setLevel(file_log_level)
        formatter = logging.Formatter(
            fmt="%(asctime)s %(name)s %(levelname)4s:  %(message)s",
            datefmt="%Y-%m-%d %H:%M:%S",
        )
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)

    return logger



# #
# @copyright Copyright (C) 2023 SpyroSoft Solutions S.A. All rights reserved.
# #

from typing import List

from scargo.config import Config
from scargo.file_generators.base_gen import create_file_from_template


def generate_conanfile(config: Config) -> None:
    create_file_from_template(
        "conan/conanfile.py.j2",
        "conanfile.py",
        template_params={"config": config},
        config=config,
    )
    create_file_from_template(
        "conan/conanfiletest.j2",
        "tests/conanfile.py",
        template_params={"config": config},
        config=config,
    )


def generate_conanprofile(config: Config) -> None:
    profiles = config.profiles.keys()
    standard_profiles: List[str] = ["Debug", "Release", "RelWithDebInfo", "MinSizeRel"]

    if config.project.target.family == "stm32":
        create_file_from_template(
            "conan/stm32_gcc_toolchain_wrapper.cmake.j2",
            ".conan/profiles/stm32_gcc_toolchain_wrapper.cmake",
            template_params={},
            config=config,
        )

    for profile in profiles:
        create_file_from_template(
            "conan/profile.j2",
            f".conan/profiles/{config.project.target.family}_{profile}",
            template_params={
                "config": config,
                "profile": profile,
                "standard_profiles": standard_profiles,
            },
            config=config,
        )


def generate_test_package(config: Config) -> None:
    create_file_from_template(
        "conan/test_package/CMakeLists.txt.j2",
        "test_package/CMakeLists.txt",
        template_params={"config": config},
        config=config,
    )

    create_file_from_template(
        "conan/test_package/conanfile.py.j2",
        "test_package/conanfile.py",
        template_params={"config": config},
        config=config,
    )

    create_file_from_template(
        "conan/test_package/example.cpp.j2",
        "test_package/src/example.cpp",
        template_params={"config": config},
        config=config,
    )



from pathlib import Path
from typing import Iterable, Tuple

from clang import native  # type: ignore[attr-defined]
from clang.cindex import Config, Index, TokenKind

Config().set_library_file(str(Path(native.__file__).with_name("libclang.so")))


def get_comment_lines(file_path: Path) -> Iterable[Tuple[int, str]]:
    index = Index.create()
    translation_unit = index.parse(str(file_path), args=["-x", "c++"])

    for token in translation_unit.cursor.get_tokens():
        if token.kind == TokenKind.COMMENT:  # pylint: disable=no-member
            yield from enumerate(
                token.spelling.splitlines(keepends=False),
                start=token.extent.start.line,
            )



# #
# @copyright Copyright (C) 2023 SpyroSoft Solutions S.A. All rights reserved.
# #

"""Build project bin/lib exec file"""
import subprocess
import sys
from pathlib import Path

from scargo.commands.publish import (
    conan_add_conancenter,
    conan_add_remote,
    conan_clean_remote,
)
from scargo.config_utils import prepare_config
from scargo.logger import get_logger

logger = get_logger()


def scargo_build(profile: str) -> None:
    """
    Build project exec file.

    :param str profile: Profile
    :return: None
    """
    config = prepare_config()

    project_dir = config.project_root
    if not project_dir:
        logger.error("Current working directory is not part of scargo project.")
        sys.exit(1)

    if not Path(project_dir, "CMakeLists.txt").exists():
        logger.error("File `CMakeLists.txt` does not exist.")
        logger.info("Did you run `scargo update`?")
        sys.exit(1)

    build_dir = Path(project_dir, "build", profile)
    build_dir.mkdir(parents=True, exist_ok=True)

    conan_clean_remote()

    conan_add_remote(project_dir, config)
    conan_add_conancenter()

    try:
        subprocess.check_call(
            [
                "conan",
                "install",
                ".",
                "-if",
                build_dir,
                "-pr:b",
                "default",
                "-pr:h",
                f"./.conan/profiles/{config.project.target.family}_{profile}",
                "-b",
                "missing",
            ],
            cwd=project_dir,
        )
        subprocess.check_call(
            [
                "conan",
                "build",
                f"{project_dir}",
            ],
            cwd=build_dir,
        )

    except subprocess.CalledProcessError:
        logger.error("Unable to build exec file")
        sys.exit(1)



# #
# @copyright Copyright (C) 2023 SpyroSoft Solutions S.A. All rights reserved.
# #

"""Clean project from unnecessary files"""
import shutil
from pathlib import Path
from typing import Optional

from scargo.config_utils import get_scargo_config_or_exit
from scargo.logger import get_logger

logger = get_logger()


def _case_insensitive_find_dir(source_dir: Path, dirname: str) -> Optional[Path]:
    if source_dir and source_dir.exists():
        for child in source_dir.iterdir():
            if child.name.lower() == dirname.lower():
                return child
    return None


def scargo_clean() -> None:
    """Clean project dir from unnecessary files"""

    config = get_scargo_config_or_exit()
    project_path = config.project_root
    test_dir = _case_insensitive_find_dir(project_path, "test")
    source_directories = [project_path, test_dir]

    for source_dir in source_directories:
        if source_dir:
            build_dir = _case_insensitive_find_dir(source_dir, "build")
            if build_dir and build_dir.exists():
                shutil.rmtree(build_dir)
                logger.info("Removed %s", build_dir)
